from . import preprocessing, train, plots, results


class Main:
    def __init__(self, ticker):
        self.ticker = ticker
        self.data_preprocessor = preprocessing.data_preprocessing(ticker)
        self.df_monthly = self.data_preprocessor.df_monthly
        print(self.data_preprocessor.data_reader.end_date)
        df_monthly = self.data_preprocessor.df_monthly
        # X, y = self.data_preprocessor.data_scaling(df_monthly)
        self.train = train.Training(ticker)
        self.train.train_model()
        self.plots = plots.Plots(ticker)
        self.plots.comparison_plots()
        self.results = results.Results(ticker)
        self.result = self.results.result
        self.result_json = self.result.to_json(orient="split")
        print(self.result)
        print(self.result_json)



import pandas as pd
import numpy as np

class Calculations():
    def __init__(self):
        self.capital_gains_tax = 0.10
        self.broker_comission = 0.003

    def gross_yield(self, df, v):
        prod = (v*df["Quotient"]+1-v).prod()
        n_years = len(v)/12
        return (prod-1)*100, ((prod**(1/n_years))-1)*100

    def separate_ones(self, u):
        u_ = np.r_[0, u, 0]
        i = np.flatnonzero(u_[:-1] != u_[1:])
        v, w = i[::2], i[1::2]
        if len(v) == 0:
            return np.zeros(len(u)), 0

        n, m = len(v), len(u)
        o = np.zeros(n*m, dtype=int)

        r = np.arange(n)*m
        o[v+r] = 1

        if w[-1] == m:
            o[w[:-1]+r[:-1]] = -1
        else:
            o[w+r] -= 1

        out = o.cumsum().reshape(n, -1)
        return out, n

    def net_yield(self, df, v):
        n_years = len(v)/12
        w, n = self.separate_ones(v)
        #print((w*np.array(df["Quotient"])+(1-w)).shape)
        A = (w*np.array(df["Quotient"])+(1-w)).prod()
        A1p = np.maximum(0, np.sign(A-1))
        Ap = A*A1p
        Am = A-Ap
        An = Am+(Ap-A1p)*(1-self.capital_gains_tax)+A1p
        prod = An.prod()*((1-self.broker_comission)**(2*n))
        return (prod-1)*100, ((prod**(1/n_years))-1)*100

    def gross_portfolio(self, df, w):
        portfolio = [(w*df["Quotient"]+(1-w))[:i].prod() for i in range(len(w))]
        return portfolio



import datetime
import yfinance as yf
import tensorflow as tf
# Restricting the GPU from consuming all memory
gpus = tf.config.experimental.list_physical_devices('GPU')
if len(gpus) > 0:
    tf.config.experimental.set_memory_growth(gpus[0], True)
    print("Using GPU")

class Data_Loader():
    def __init__(self, ticker):
        self.ticker = ticker
        self.df = self.load_data()
        self.start_date = self.df.index[0].to_pydatetime()
        self.end_date = datetime.date.today()
        self.first_days = self.date_calculations()

    def load_data(self):
        stock = yf.Ticker(self.ticker)
        stock.info
        # get historical market data
        df = stock.history(period="max")
        df.drop("Dividends", axis=1, inplace=True)
        df.drop("Stock Splits", axis=1, inplace=True)
        return df
    
    def date_calculations(self):
        start_year = self.start_date.year
        start_month = self.start_date.month
        end_year = self.end_date.year
        end_month = self.end_date.month

        first_days = []  

        for month in range(start_month, 13):
            first_days.append(min(self.df.loc[str(start_year)+"-"+str(month)].index))

        for year in range(start_year+1, end_year):
            for month in range(1, 13):
                first_days.append(min(self.df.loc[str(year)+"-"+str(month)].index))
        
        for month in range(1, end_month+1):
            first_days.append(min(self.df.loc[str(end_year)+"-"+str(month)].index))

        return first_days


import numpy as np
from . import train
class Market():
    def __init__(self, ticker):
        self.train = train.Training(ticker)
        self.w_lstm = np.diff(self.train.y_pred_lstm.reshape(self.train.y_pred_lstm.shape[0]), 1)
        self.v_lstm = np.maximum(np.sign(self.w_lstm), 0)
        self.w_mix = np.diff(self.train.y_pred_mix.reshape(self.train.y_pred_mix.shape[0]), 1)
        self.v_mix = np.maximum(np.sign(self.w_mix), 0)


import matplotlib.pyplot as plt
from . import train, preprocessing, market, calculations


class Plots:
    def __init__(self, ticker):
        self.preprocessing = preprocessing.data_preprocessing(ticker)
        self.train = train.Training(ticker)
        # self.training_variables = training_variables.Training_Variables(ticker)
        self.market = market.Market(ticker)
        self.calculations = calculations.Calculations()

    def plot_training_data(self, model_selected, metric='loss', val_metric='val_loss'):
        plt.plot(model_selected.history[metric])
        plt.plot(model_selected.history[val_metric])
        plt.title('model loss')
        plt.ylabel('loss')
        plt.xlabel('epoch')
        plt.legend(['train', 'test'], loc='upper right')
        plt.show()
    
    def plot_predictions(self):
        plt.figure(figsize=(30, 10))
        plt.plot(self.preprocessing.y_train, label="Actual")
        plt.plot(self.train.y_pred_train_lstm, label="Prediction by LSTM Model")
        plt.plot(self.train.y_pred_train_mix, label="Prediction by Mix-LSTM Model")
        plt.legend(fontsize=20)
        plt.grid(axis="both")
        plt.title("Actual Open Price and Pedicted Ones on train set", fontsize=20)
        plt.show()

    def in_out(self):
        plt.figure(figsize=(30, 10))
        plt.plot(self.preprocessing.y_test, label="Actual")
        plt.plot(self.train.y_pred_lstm, label="LSTM Predictions")
        plt.plot(self.market.v_lstm, label="In and out LSTM")
        plt.plot(self.train.y_pred_mix, label="Mix LSTM Predictions")
        plt.plot(self.market.v_mix, label="In and out Mix LSTM")
        plt.legend(fontsize=20)
        plt.grid(axis="both")
        plt.title(
            "Actual Open Price, Predicted Ones and Vectors on In and Out Moments", fontsize=20)
        plt.show()

    def comparison_plots(self):
        plt.figure(figsize=(30, 10))
        plt.plot(self.calculations.gross_portfolio(self.preprocessing.test, self.preprocessing.v_bh), label="Portfolio Buy and Hold Strategy")
        plt.plot(self.calculations.gross_portfolio(self.preprocessing.test, self.preprocessing.v_ma),
                label="Portfolio Moving Average Strategy")
        plt.plot(self.calculations.gross_portfolio(self.preprocessing.test, self.market.v_lstm), label="Portfolio LSTM Model")
        plt.plot(self.calculations.gross_portfolio(self.preprocessing.test, self.market.v_mix), label="Portfolio Mix LSTM Model")
        plt.legend(fontsize=20)
        plt.grid(axis="both")
        plt.title("Gross Portfolios of three methods", fontsize=20)
        plt.show()



from keras.callbacks import TensorBoard, ReduceLROnPlateau
from . import preprocessing
from . import models


class Training:
    def __init__(self, ticker):

        self.preprocessing = preprocessing.data_preprocessing(ticker)
        self.models = models.Models(ticker)
        self.learning_rate_reduction = ReduceLROnPlateau(monitor='val_loss',
                                            patience=25,
                                            verbose=1,
                                            factor=0.25,
                                            min_lr=0.00001)
        """self.tensorboard = keras.callbacks.TensorBoard(
                                                        log_dir='TensorBoard_Logs',
                                                        histogram_freq=1,
                                                        embeddings_freq=1
                                                    )  
        """
        # self.lstm_history, self.mix_history = self.train_model()
        # self.models.lstm_model.save_weights(lstm_weights.h5")
        # self.models.mix_lstm_model.save_weights("mix_lstm_weights.h5")
        self.y_pred_train_lstm = self.models.lstm_model.predict(self.preprocessing.X_train)
        self.y_pred_train_mix = self.models.mix_lstm_model.predict(self.preprocessing.X_train)
        self.y_pred_lstm = self.models.lstm_model.predict(self.preprocessing.X_test)
        self.y_pred_mix = self.models.mix_lstm_model.predict(self.preprocessing.X_test)

    def train_model(self):
        lstm_history = self.models.lstm_model.fit(self.preprocessing.X_train, self.preprocessing.y_train, epochs=400, batch_size=48, validation_data=(self.preprocessing.X_test, self.preprocessing.y_test),
                              verbose=1, callbacks=[self.learning_rate_reduction], shuffle=False)

        mix_history = self.models.mix_lstm_model.fit(self.preprocessing.X_train, self.preprocessing.y_train, epochs=400, batch_size=48, validation_data=(self.preprocessing.X_test, self.preprocessing.y_test),
                                 verbose=1, callbacks=[self.learning_rate_reduction], shuffle=False)
        return lstm_history, mix_history




import pandas as pd
from . import calculations, preprocessing, market


class Results:
    def __init__(self, ticker):
        self.preprocessing = preprocessing.data_preprocessing(ticker)
        self.market = market.Market(ticker)
        self.calculations = calculations.Calculations()
        self.result = self.result_calculations()

    def result_calculations(self):
        print("Test period of {:.2f} years, from {} to {} \n".format(len(self.preprocessing.v_bh)/12, 
                str(self.preprocessing.test.loc[self.preprocessing.test.index[0], "First Day Current Month"])[:10]
                , str(self.preprocessing.test.loc[self.preprocessing.test.index[-1], "First Day Next Month"])[:10]))

        results = pd.DataFrame({})
        results["Method"] = ["Buy and hold", "Moving average", "LSTM", "Mix"]
        vs = [self.preprocessing.v_bh, self.preprocessing.v_ma, self.market.v_lstm, self.market.v_mix]
        # print(vs)
        results["Total Gross Yield"] = [str(round(self.calculations.gross_yield(self.preprocessing.test, vi)[0], 2))+" %" for vi in vs]
        results["Annual Gross Yield"] = [str(round(self.calculations.gross_yield(self.preprocessing.test, vi)[1], 2))+" %" for vi in vs]
        results["Total Net Yield"] = [str(round(self.calculations.net_yield(self.preprocessing.test, vi)[0], 2))+" %" for vi in vs]
        results["Annual Net Yield"] = [str(round(self.calculations.net_yield(self.preprocessing.test, vi)[1], 2))+" %" for vi in vs]
        
        return results
        


from keras.models import Sequential
from keras.optimizers import Adam
from keras.layers import Dense, Dropout, Conv1D, LSTM

from . import preprocessing


class Models():
    def __init__(self, ticker):
        self.preprocessing = preprocessing.data_preprocessing(ticker)
        self.lstm_model = self.LSTM_Model(self.preprocessing.window+1, 8)
        self.mix_lstm_model = self.Mix_LSTM_Model(self.preprocessing.window+1, 8)

    def LSTM_Model(self, window, features):
        model = Sequential()
        model.add(LSTM(300, input_shape=(window, features), return_sequences=True))
        model.add(Dropout(0.5))
        model.add(LSTM(200,  return_sequences=False))
        model.add(Dropout(0.5))
        model.add(Dense(100, kernel_initializer='uniform', activation='relu'))
        model.add(Dense(1, kernel_initializer='uniform', activation='relu'))
        model.compile(loss='mse', optimizer=Adam(lr=0.001))

        return model

    def Mix_LSTM_Model(self, window, features):

        model = Sequential()
        model.add(Conv1D(input_shape=(window, features), filters=32,
                        kernel_size=2, strides=1, activation='relu', padding='same'))
        model.add(Conv1D(filters=64, kernel_size=2, strides=1,
                        activation='relu', padding='same'))
        model.add(LSTM(300, return_sequences=True))
        model.add(Dropout(0.5))
        model.add(LSTM(200,  return_sequences=False))
        model.add(Dropout(0.5))
        model.add(Dense(100, kernel_initializer='uniform', activation='relu'))
        model.add(Dense(1, kernel_initializer='uniform', activation='relu'))

        model.compile(loss='mse', optimizer=Adam(lr=0.001))

        return model



import numpy as np
from sklearn.preprocessing import MinMaxScaler
import pandas as pd
from . import data


class data_preprocessing:
    def __init__(self, ticker):
        self.data_reader = data.Data_Loader(ticker)
        self.df_monthly = self.monthly_df(self.data_reader.df)
        self.window = 5
        self.X, self.y = self.data_scaling(self.df_monthly)
        self.split = 72
        self.X_train = self.X[:-self.split-1, :, :]
        self.X_test = self.X[-self.split-1:, :, :]
        self.y_train = self.y[:-self.split-1]
        self.y_test = self.y[-self.split-1:]
        self.test = self.df_monthly.iloc[-self.split:, :]
        self.v_bh = np.ones(self.test.shape[0])
        self.v_ma = self.test["First Day Current Month Opening"] > self.test["mv_avg_12"]

    def monthly_df(self, df):

        dfm = df.resample("M").mean()
        dfm = dfm[:-1]

        dfm["First Day Current Month"] = self.data_reader.first_days[:-1]
        dfm["First Day Next Month"] = self.data_reader.first_days[1:]
        dfm["First Day Current Month Opening"] = np.array(
            df.loc[self.data_reader.first_days[:-1], "Open"])
        dfm["First Day Next Month Opening"] = np.array(
            df.loc[self.data_reader.first_days[1:], "Open"])
        dfm["Quotient"] = dfm["First Day Next Month Opening"].divide(
            dfm["First Day Current Month Opening"])

        dfm["mv_avg_12"] = dfm["Open"].rolling(window=12).mean().shift(1)
        dfm["mv_avg_24"] = dfm["Open"].rolling(window=24).mean().shift(1)

        dfm = dfm.iloc[24:, :]
        return dfm

    def create_window(self, data, window_size):
        data_s = data.copy()
        for i in range(window_size):
            data = pd.concat([data, data_s.shift(-(i + 1))], axis=1)

        data.dropna(axis=0, inplace=True)
        return(data)

    def data_scaling(self, dfm):
        scaler = MinMaxScaler(feature_range=(0, 1))
        dg = pd.DataFrame(scaler.fit_transform(dfm[["High", "Low", "Open", "Close", "Volume", "First Day Current Month Opening",
                                                    "mv_avg_12", "mv_avg_24", "First Day Next Month Opening"]].values))
        X = dg[[0, 1, 2, 3, 4, 5, 6, 7]]
        X = self.create_window(X, self.window)
        X = np.reshape(X.values, (X.shape[0], self.window+1, 8))
        y = np.array(dg[8][self.window:])

        return X, y
